from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import is_matched_canonical, filter_tag


class JuejinExtractor(Extractor):
    """
    juejin.cn
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return is_matched_canonical("https://juejin.cn/", soup)

    @override
    def article_container(self) -> tuple:
        return ("div", {"id": "article-root"})

    @override
    def extract_title(self, soup: BeautifulSoup) -> str:
        title_tag = filter_tag(soup.find("h1", {"class": "article-title"}))
        return title_tag.get_text(strip=True) if title_tag else super().extract_title(soup)



from typing import override
from bs4 import BeautifulSoup
from bs4.element import Tag

from ..extractor import Extractor
from ..utils import filter_tag, get_attr_text, get_og_site_name


class WechatGZHExtractor(Extractor):
    """
    微信公众号
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.append(lambda el: 'id' in el.attrs and el.attrs['id'] == 'meta_content')

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_site_name(soup) == "微信公众平台"

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "rich_media_content"})

    @override
    def extract_img(self, element: Tag) -> Tag:
        img_els = element.find_all("img")
        for img_el in img_els:
            img_tag = filter_tag(img_el)
            if img_tag:
                src = get_attr_text(img_tag.attrs.get("data-src"))
                if src:
                    img_tag.attrs["src"] = src
        return element



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import is_matched_canonical


class InfoQCNExtractor(Extractor):
    """
    www.infoq.cn
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return is_matched_canonical("https://www.infoq.cn", soup)

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "article-content-wrap"})



import json
from pathlib import Path
from typing import Any


class Store:
    def __init__(self, base_dir_name: str = ".ommimd"):
        self.path = Path.home() / base_dir_name

    def save(self, key: str, obj: Any):
        self.path.mkdir(parents=True, exist_ok=True)
        file_path = self.path / f"{key}.json"
        with open(file_path, "w", encoding="utf8") as f:
            json.dump(obj, f, indent=4, ensure_ascii=False)

    def load(self, key: str) -> Any | None:
        file_path = self.path / f"{key}.json"
        if not file_path.exists() or not file_path.is_file():
            return None
        with open(file_path, "r", encoding="utf8") as f:
            return json.load(f)



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_title


class ClaudeDocExtractor(Extractor):
    """
    docs.claude.com
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend(
            [
                lambda el: "data-component-part" in el.attrs and "code-block-header" in el.attrs["data-component-part"],
                lambda el: "data-component-part" in el.attrs and "code-group-tab-bar" in el.attrs["data-component-part"],
            ]
        )

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_title(soup).endswith(" - Claude Docs")

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "mdx-content"})



import json
import re
from typing import override
from bs4 import BeautifulSoup
import requests

from ..extractor import Article, Extractor
from ..utils import REQUEST_HEADERS, filter_tag, get_og_url


class YuqueExtractor(Extractor):
    """
    语雀
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_url(soup).startswith("https://www.yuque.com")

    @override
    def article_container(self) -> tuple:
        return ("", {})

    @override
    def extract_article(self, soup: BeautifulSoup) -> Article:
        script_tag = filter_tag(soup.find("script", string=re.compile(r"decodeURIComponent")))
        if script_tag:
            raw_js = script_tag.string
            if raw_js:
                match = re.search(r'decodeURIComponent\s*\(\s*"([^"]+)"\s*\)', raw_js)
                if match:
                    encoded_str = match.group(1)

                    from urllib.parse import unquote

                    decoded_str = unquote(encoded_str)
                    decoded_json = json.loads(decoded_str)
                    # print(decoded_json)
                    doc = decoded_json["doc"]
                    if doc and doc["book_id"]:
                        book_id = str(doc["book_id"])
                        slug = str(doc["slug"])
                        response = requests.get(f"https://www.yuque.com/api/docs/{slug}?book_id={book_id}&mode=markdown", headers=REQUEST_HEADERS)
                        response.encoding = "utf-8"
                        resp = response.json()
                        # print(resp)
                        return Article(str(resp["data"]["title"]), None, None, str(resp["data"]["sourcecode"]))
        return Article("", None, None, "")



from abc import ABC, abstractmethod
from dataclasses import dataclass
import re
from typing import Callable, override

from bs4 import BeautifulSoup
from bs4.element import Tag, Comment

from .utils import filter_tag, get_attr_text, get_canonical_url, get_og_description, get_og_title, get_og_url, get_title


TAGS_TO_CLEAN: list[Callable[[Tag], bool]] = [
    lambda el: el.name in ("style", "link", "button", "footer", "header", "aside"),
    lambda el: el.name == "script" and "src" not in el.attrs,
    lambda el: el.name == "script"
        and el.has_attr("src")
        and not get_attr_text(el.attrs["src"]).startswith("https://gist.github.com"),
]

ATTRS_TO_CLEAN: list[Callable[[Tag], bool]] = [
    lambda el: "style" in el.attrs
        and re.search(r"display\s*:\s*none", get_attr_text(el.attrs.get("style")), re.IGNORECASE) is not None,
    lambda el: "hidden" in el.attrs,
    lambda el: "class" in el.attrs and "katex-html" in el.attrs["class"],  # katex
]

ARTICLE_CONTAINERS = [("article", None), ("main", None), ("body", None)]


@dataclass
class Article:
    title: str
    url: str | None
    description: str | None
    body: Tag | str


class Extractor(ABC):
    def __init__(self):
        self.tags_to_clean = TAGS_TO_CLEAN
        self.attrs_to_clean = ATTRS_TO_CLEAN

    def extract(self, soup: BeautifulSoup) -> Article | None:
        if self.can_handle(soup):
            # print(f"Using extractor: {self.__class__.__name__}")
            self.pre_handle_soup(soup)
            article_container = self.article_container()
            if isinstance(article_container, tuple):
                article_container = [article_container]
            for container in article_container:
                article = self.extract_article(soup)
                if article:
                    return article
                article_tag = extract_article_from_soup(soup, container)
                if article_tag:
                    for el in article_tag.find_all():
                        tag = filter_tag(el)
                        if tag:
                            if any(cond(tag) for cond in self.tags_to_clean):
                                tag.decompose()
                                continue
                            if tag.attrs:
                                if any(cond(tag) for cond in self.attrs_to_clean):
                                    tag.decompose()
                    for comment in article_tag.find_all(string=lambda text: isinstance(text, Comment)):
                        comment.extract()
                    self.extract_img(article_tag)
                    title = self.extract_title(soup)
                    description = self.extract_description(soup)
                    url = self.extract_url(soup)
                    article = Article(title=title, url=url, description=description, body=article_tag)
                    remove_duplicate_titles(article)
                    return article
        return None

    @abstractmethod
    def can_handle(self, soup: BeautifulSoup) -> bool: ...

    @abstractmethod
    def article_container(self) -> tuple | list: ...

    def extract_title(self, soup: BeautifulSoup) -> str:
        return get_og_title(soup) or get_title(soup)

    def extract_description(self, soup: BeautifulSoup) -> str:
        return get_og_description(soup)

    def extract_url(self, soup: BeautifulSoup) -> str:
        return get_og_url(soup) or get_canonical_url(soup)

    def extract_img(self, element: Tag) -> Tag:
        return element

    def extract_article(self, soup: BeautifulSoup) -> Article | None:
        return None

    def pre_handle_soup(self, soup: BeautifulSoup):
        ...


class DefaultExtractor(Extractor):
    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return True

    @override
    def article_container(self) -> tuple | list:
        return ARTICLE_CONTAINERS


def extract_article_from_soup(soup: BeautifulSoup, template: tuple) -> Tag | None:
    if template[1] is not None:
        result = soup.find(template[0], attrs=template[1])
    else:
        result = soup.find(template[0])
    return filter_tag(result)


def remove_duplicate_titles(article: Article):
    if article.body and isinstance(article.body, Tag):
        first_h1 = article.body.find("h1")
        if first_h1:
            h1_text = first_h1.get_text(strip=True)
            if h1_text.lower() in article.title.lower():
                article.title = h1_text
                first_h1.decompose()



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_site_name


class ZhihuExtractor(Extractor):
    """
    知乎专栏
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_site_name(soup) == "知乎专栏"

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "Post-RichText"})



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_url


class MicrosoftLearnExtractor(Extractor):
    """
    微软技术文档
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend(
            [
                lambda el: "id" in el.attrs and "article-header" in el.attrs["id"],
                lambda el: "id" in el.attrs and "article-metadata" in el.attrs["id"],
                lambda el: "id" in el.attrs and "site-user-feedback-footer" in el.attrs["id"],
            ]
        )

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_url(soup).startswith("https://learn.microsoft.com")

    @override
    def article_container(self) -> tuple:
        return ("main", {"id": "main"})



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_site_name


class TowardsDataScienceExtractor(Extractor):
    """
    towardsdatascience.com
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend([
            lambda el: 'class' in el.attrs and 'taxonomy-post_tag' in el.attrs['class'],
            lambda el: 'class' in el.attrs and 'tds-cta-box' in el.attrs['class'],
            lambda el: 'class' in el.attrs and 'wp-block-buttons' in el.attrs['class'],
            lambda el: 'class' in el.attrs and 'wp-block-outermost-social-sharing' in el.attrs['class'],
            lambda el: 'class' in el.attrs and 'wp-block-tenup-post-time-to-read' in el.attrs['class'],
        ])
        self.tags_to_clean.extend([
            lambda el: el.name == 'time',
        ])

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_site_name(soup) == "Towards Data Science"

    @override
    def article_container(self) -> tuple | list:
        return ("main", None)



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor


class TencentCloudExtractor(Extractor):
    """
    腾讯云开发者社区
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        title_tag = soup.title
        title = title_tag.get_text(strip=True) if title_tag else None
        return title is not None and title.endswith("-腾讯云开发者社区-腾讯云")

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "mod-content__markdown"})



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_url


class WoShiPMExtractor(Extractor):
    """
    人人都是产品经理
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_url(soup).startswith("https://www.woshipm.com")

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "article--content"})



import importlib
import pkgutil
from dataclasses import dataclass
from pathlib import Path

from bs4 import BeautifulSoup

from .extractor import Article, DefaultExtractor, Extractor
from .parser import HtmlMarkdownParser
from .readers import ReaderFactory
from .utils import to_snake_case


@dataclass
class ReaderContext:
    raw_html: str


@dataclass
class ExtractorContext:
    article: Article


@dataclass
class ParserContext:
    title: str
    markdown: str


class OmniArticleMarkdown:
    DEFAULT_SAVE_PATH = "./"

    def __init__(self, url_or_path: str):
        self.url_or_path = url_or_path

    def parse(self) -> ParserContext:
        reader_ctx = self._read_html(self.url_or_path)
        extractor_ctx = self._extract_article(reader_ctx)
        parser_ctx = self._parse_html(extractor_ctx)
        return parser_ctx

    def save(self, ctx: ParserContext, save_path: str = "") -> str:
        save_path = save_path or self.DEFAULT_SAVE_PATH
        file_path = Path(save_path)
        if file_path.is_dir():
            filename = f"{to_snake_case(ctx.title)}.md"
            file_path = file_path / filename
        with file_path.open("w", encoding="utf-8") as f:
            f.write(ctx.markdown)
        return str(file_path.resolve())

    def _read_html(self, url_or_path: str) -> ReaderContext:
        reader = ReaderFactory.create(url_or_path)
        raw_html = reader.read()
        return ReaderContext(raw_html)

    def _extract_article(self, ctx: ReaderContext) -> ExtractorContext:
        soup = BeautifulSoup(ctx.raw_html, "html5lib")
        for extract in load_extractors():
            article = extract.extract(soup)
            if article:
                break
        else:
            article = DefaultExtractor().extract(soup)
        if not article:
            raise ValueError("Failed to extract article content.")
        return ExtractorContext(article)

    def _parse_html(self, ctx: ExtractorContext) -> ParserContext:
        parser = HtmlMarkdownParser(ctx.article)
        result = parser.parse()
        return ParserContext(title=result[0], markdown=result[1])


def load_extractors(package_name="extractors") -> list[Extractor]:
    extractors_package = Path(__file__).parent / package_name
    extractors = []
    for loader, module_name, is_pkg in pkgutil.iter_modules([extractors_package.resolve()]):
        module = importlib.import_module(f"omni_article_markdown.{package_name}.{module_name}")
        for attr in dir(module):
            cls = getattr(module, attr)
            if isinstance(cls, type) and issubclass(cls, Extractor) and cls is not Extractor:
                extractors.append(cls())
    return extractors



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import filter_tag


class Netease163Extractor(Extractor):
    """
    百家号
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        tag1 = filter_tag(soup.find("div", {"data-testid": "article"}))
        tag2 = filter_tag(soup.find("span", {"class": "bjh-p"}))
        return tag1 is not None and tag2 is not None

    @override
    def article_container(self) -> tuple:
        return ("div", {"data-testid": "article"})

    @override
    def pre_handle_soup(self, soup: BeautifulSoup):
        for tag in soup.find_all("span", {"class": "bjh-p"}):
            span_tag = filter_tag(tag)
            if span_tag:
                span_tag.name = "p"
        # for tag in soup.find_all("img"):
        #     tag.wrap(soup.new_tag("p"))



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_site_name


class AppleDevelopExtractor(Extractor):
    """
    Apple Developer Documentation
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend(
            [
                lambda el: "class" in el.attrs and "eyebrow" in el.attrs["class"],
                lambda el: "class" in el.attrs and "platform" in el.attrs["class"],
                lambda el: "class" in el.attrs and "title" in el.attrs["class"],
            ]
        )

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_site_name(soup) == "Apple Developer Documentation"

    @override
    def article_container(self) -> tuple:
        return ("main", {"class": "main"})



import click
import sys
from runpy import run_module
from click_default_group import DefaultGroup

from .omni_article_md import OmniArticleMarkdown
from . import DEFAULT_PLUGINS


@click.group(cls=DefaultGroup, default="parse", default_if_no_args=True)
def cli():
    """
    A CLI tool to parse articles and save them as Markdown.
    It also supports installing plugins.
    """
    ...


@cli.command(name="parse")
@click.argument("url_or_path")
@click.option(
    "-s",
    "--save",
    help="Save result (default: ./). Provide a path to save elsewhere.",
    type=click.Path(dir_okay=True, writable=True),
)
def parse_article(url_or_path: str, save: str | None):
    """
    Parses an article from a URL or local path and outputs/saves it as Markdown.
    """
    handler = OmniArticleMarkdown(url_or_path)
    parser_ctx = handler.parse()

    if save is None:
        click.echo(parser_ctx.markdown)
    else:
        save_path = handler.save(parser_ctx, save)
        click.echo(f"Article saved to: {save_path}")


@cli.command()
@click.argument("plugin_name")
@click.option("-U", "--upgrade", is_flag=True, help="Upgrade the plugin if already installed.", default=False)
@click.option(
    "-e",
    "--editable",
    is_flag=True,
    help="Install the editable package based on the provided local file path",
    default=False,
)
def install(plugin_name: str, upgrade: bool, editable: bool):
    """
    Installs a plugin for this application.
    For example, to install the 'zhihu' plugin: mdcli install zhihu
    """
    actual_package_name = (
        plugin_name if editable or plugin_name not in DEFAULT_PLUGINS else DEFAULT_PLUGINS[plugin_name]
    )

    click.echo(f"Attempting to install plugin: {actual_package_name}...")
    args = ["pip", "install"]
    if upgrade:
        args.append("--upgrade")
    args.append(actual_package_name)

    original_argv = sys.argv
    try:
        sys.argv = args
        run_module("pip", run_name="__main__")
        click.echo(f"Plugin '{actual_package_name}' processed by pip.")
        click.echo("If the plugin provides new functionality, it should now be available.")
        click.echo(
            "You might need to restart the application for changes to take full effect if it involves runtime loading during startup."
        )
    except Exception as e:
        click.echo(f"Failed to process plugin '{actual_package_name}' with pip: {e}", err=True)
        click.echo("Please ensure pip is installed and the package name is correct.", err=True)
    finally:
        sys.argv = original_argv


@cli.command()
@click.argument("plugin_name")
@click.option("-y", "--yes", is_flag=True, help="Don't ask for confirmation before uninstalling.", default=False)
def uninstall(plugin_name: str, yes: bool):
    """
    Uninstalls a plugin for this application.
    For example, to uninstall the 'zhihu' plugin: mdcli uninstall zhihu
    """
    actual_package_name = plugin_name if plugin_name not in DEFAULT_PLUGINS else DEFAULT_PLUGINS[plugin_name]

    click.echo(f"Attempting to uninstall plugin: {actual_package_name}...")
    args = ["pip", "uninstall"]
    if yes:
        args.append("-y")
    args.append(actual_package_name)

    original_argv = sys.argv
    try:
        sys.argv = args
        run_module("pip", run_name="__main__")
        click.echo(f"Plugin '{actual_package_name}' uninstallation processed by pip.")
        click.echo(
            "The plugin's functionality should no longer be available after the next application start (or if dynamically unloaded)."
        )
    except Exception as e:
        click.echo(f"Failed to process uninstallation of plugin '{actual_package_name}' with pip: {e}", err=True)
        click.echo("Please ensure pip is installed and the package name is correct.", err=True)
    finally:
        sys.argv = original_argv


if __name__ == "__main__":
    cli()



from typing import Protocol

from pluggy import HookimplMarker, HookspecMarker

hookspec = HookspecMarker("mdcli")
hookimpl = HookimplMarker("mdcli")


class ReaderPlugin(Protocol):
    def can_handle(self, url: str) -> bool: ...

    def read(self, url: str) -> str: ...


@hookspec(firstresult=True)
def get_custom_reader(url: str) -> ReaderPlugin | None:
    """
    Allows plugins to provide a custom reader for a given URL.
    The first plugin that returns a ReaderPlugin instance will be used.
    """
    ...



import re
from urllib.parse import urlparse

from bs4 import BeautifulSoup
from bs4.element import NavigableString, Tag, PageElement, AttributeValueList

REQUEST_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36 Edg/126.0.0.0",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8,en-GB;q=0.7,en-US;q=0.6",
    "Priority": "u=0, i",
    "Sec-Ch-Ua": '"Not/A)Brand";v="8", "Chromium";v="126", "Microsoft Edge";v="126"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": '"macOS"',
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "none",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
}

BROWSER_TARGET_HOSTS = [
    "developer.apple.com/documentation/",
    "www.infoq.cn/",
    "pcsx2.net/",
    "baijiahao.baidu.com/",
]

def is_sequentially_increasing(code: str) -> bool:
    try:
        # 解码并按换行符拆分
        numbers = [int(line.strip()) for line in code.split("\n") if line.strip()]
        # 检查是否递增
        return all(numbers[i] + 1 == numbers[i + 1] for i in range(len(numbers) - 1))
    except ValueError:
        return False  # 处理非数字情况


def move_spaces(input_string: str, suffix: str) -> str:
    # 使用正则表达式匹配以指定的suffix结尾，且suffix之前有空格的情况
    escaped_suffix = re.escape(suffix)  # 处理正则中的特殊字符
    pattern = rf"(.*?)\s+({escaped_suffix})$"
    match = re.search(pattern, input_string)
    if match:
        # 获取字符串的主体部分（不含空格）和尾部的 '**'
        main_part = match.group(1)
        stars = match.group(2)
        # 计算空格的数量并将空格移动到 '**' 后
        space_count = len(input_string) - len(main_part) - len(stars)
        return f"{main_part}{stars}{' ' * space_count}"
    return input_string


def to_snake_case(input_string: str) -> str:
    input_string = "".join(char if char.isalnum() else " " for char in input_string)
    snake_case_string = "_".join(word.lower() for word in input_string.split())
    return snake_case_string


def collapse_spaces(text) -> str:
    """
    将多个连续空格（包括换行和 Tab）折叠成一个空格。
    """
    return re.sub(r"\s+", " ", text)


def extract_domain(url: str) -> str | None:
    """
    从URL中提取域名（包含协议）。

    Args:
        url (str): 要提取域名的URL。

    Returns:
        str | None: 提取出的域名（包含协议），如果解析失败或协议不支持则返回 None。
    """
    try:
        parsed_url = urlparse(url)
        if parsed_url.scheme in {"http", "https"} and parsed_url.netloc:
            return f"{parsed_url.scheme}://{parsed_url.netloc}".rstrip("/")
        return None  # 返回 None 表示 URL 格式不符合要求或协议不支持

    except ValueError:
        return None  # 如果 URL 格式无效，则返回 None


def detect_language(file_name: str | None, code: str) -> str:
    # TODO: 添加语言检测逻辑
    return ""


def filter_tag(el: Tag | PageElement | NavigableString | None) -> Tag | None:
    if el is None or not isinstance(el, Tag):
        return None
    return el


def get_attr_text(el: str | AttributeValueList | None) -> str:
    if el is None:
        return ""
    if isinstance(el, str):
        return el.strip()
    else:
        return " ".join(el).strip()


def get_og_url(soup: BeautifulSoup) -> str:
    og_tag = filter_tag(soup.find("meta", {"property": "og:url"}))
    return get_tag_text(og_tag, "content")


def get_og_site_name(soup: BeautifulSoup) -> str:
    og_tag = filter_tag(soup.find("meta", {"property": "og:site_name"}))
    return get_tag_text(og_tag, "content")


def get_og_description(soup: BeautifulSoup) -> str:
    og_tag = filter_tag(soup.find("meta", {"property": "og:description"}))
    return get_tag_text(og_tag, "content")


def get_canonical_url(soup: BeautifulSoup) -> str:
    canonical_tag = filter_tag(soup.find("link", {"rel": "canonical"}))
    return get_tag_text(canonical_tag, "href")


def is_matched_canonical(url: str, soup: BeautifulSoup) -> bool:
    canonical = get_canonical_url(soup)
    if not canonical:
        return False
    return canonical.startswith(url)


def get_og_title(soup: BeautifulSoup) -> str:
    og_tag = filter_tag(soup.find("meta", {"property": "og:title"}))
    return get_tag_text(og_tag, "content")


def get_tag_text(tag: Tag | None, attr: str) -> str:
    if tag is not None and tag.has_attr(attr):
        el = tag[attr]
        return get_attr_text(el)
    return ""


def get_title(soup: BeautifulSoup) -> str:
    title_tag = soup.title
    return title_tag.get_text(strip=True) if title_tag else ""



import pluggy
from . import hookspecs

pm = pluggy.PluginManager("mdcli")
pm.add_hookspecs(hookspecs)

_loaded_plugins = False

def load_mdcli_plugins():
    global _loaded_plugins
    if _loaded_plugins:
        return
    pm.load_setuptools_entrypoints("mdcli")
    _loaded_plugins = True

# 在应用启动时调用
load_mdcli_plugins()



import re
from typing import Callable
from urllib.parse import urljoin
from bs4.element import NavigableString, Tag
import requests

from .extractor import Article
from .utils import (
    filter_tag,
    get_attr_text,
    is_sequentially_increasing,
    move_spaces,
    detect_language,
    collapse_spaces,
)

LB_SYMBOL = "[|lb_bl|]"

POST_HANDLERS: list[Callable[[str], str]] = [
    # 添加换行使文章更美观
    lambda el: re.sub(f"(?:{re.escape(LB_SYMBOL)})+", LB_SYMBOL, el).replace(LB_SYMBOL, "\n\n").strip(),
    # 纠正不规范格式 `**code**` 替换为 **`code`**
    lambda el: re.sub(r"`\*\*(.*?)\*\*`", r"**`\1`**", el),
    # 纠正不规范格式 `*code*` 替换为 *`code`*
    lambda el: re.sub(r"`\*(.*?)\*`", r"*`\1`*", el),
    # 纠正不规范格式 `[code](url)` 替换为 [`code`](url)
    lambda el: re.sub(r"`\s*\[([^\]]+)\]\(([^)]+)\)\s*`", r"[`\1`](\2)", el),
    # 将 \( ... \) 替换为 $ ... $
    lambda el: re.sub(r"\\\((.+?)\\\)", r"$\1$", el),
    # 将 \[ ... \] 替换为 $$ ... $$
    lambda el: re.sub(r"\\\[(.+?)\\\]", r"$$\1$$", el),
]

INLINE_ELEMENTS = ["span", "code", "li", "a", "strong", "em", "b", "i"]

BLOCK_ELEMENTS = [
    "p",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "ul",
    "ol",
    "blockquote",
    "pre",
    "img",
    "picture",
    "hr",
    "figcaption",
    "table",
    "section",
]

TRUSTED_ELEMENTS = INLINE_ELEMENTS + BLOCK_ELEMENTS


class HtmlMarkdownParser:
    def __init__(self, article: Article):
        self.article = article

    def parse(self) -> tuple[str, str]:
        if isinstance(self.article.body, str):
            markdown = self.article.body
        else:
            markdown = self._process_children(self.article.body)
        for handler in POST_HANDLERS:
            markdown = handler(markdown)
        if not self.article.description or self.article.description in markdown:
            description = ""
        else:
            description = f"> {self.article.description}\n\n"
        result = f"# {self.article.title}\n\n{description}{markdown}"
        # print(result)
        return (self.article.title, result)

    def _process_element(self, element: Tag, level: int = 0, is_pre: bool = False) -> str:
        parts = []
        tag = element.name
        match tag:
            case "br":
                parts.append(LB_SYMBOL)
            case "hr":
                parts.append("---")
            case "h1" | "h2" | "h3" | "h4" | "h5" | "h6":
                heading = self._process_children(element, level, is_pre=is_pre)
                parts.append(f"{'#' * int(element.name[1])} {heading}")
            case "a":
                link = self._process_children(element, level, is_pre=is_pre).replace(LB_SYMBOL, "")
                if link:
                    parts.append(f"[{link}]({element.get('href')})")
            case "strong" | "b":
                parts.append(move_spaces(f"**{self._process_children(element, level, is_pre=is_pre)}**", "**"))
            case "em" | "i":
                parts.append(move_spaces(f"*{self._process_children(element, level, is_pre=is_pre)}*", "*"))
            case "ul" | "ol":
                parts.append(self._process_list(element, level))
            case "img":
                parts.append(self._process_image(element, None))
            case "blockquote":
                blockquote = self._process_children(element, level, is_pre=is_pre)
                if blockquote.startswith(LB_SYMBOL):
                    blockquote = blockquote.removeprefix(LB_SYMBOL)
                if blockquote.endswith(LB_SYMBOL):
                    blockquote = blockquote.removesuffix(LB_SYMBOL)
                parts.append("\n".join(f"> {line}" for line in blockquote.split(LB_SYMBOL)))
            case "pre":
                parts.append(self._process_codeblock(element, level))
            case "code":  # inner code
                code = self._process_children(element, level, is_pre=is_pre)
                if LB_SYMBOL not in code:
                    parts.append(f"`{code}`")
                else:
                    parts.append(code)
            case "picture":
                source_elements = element.find_all("source")
                img_element = filter_tag(element.find("img"))
                if img_element and source_elements:
                    el = source_elements[0]
                    src_el = filter_tag(el)
                    if src_el:
                        parts.append(self._process_image(img_element, src_el))
                elif img_element:
                    parts.append(self._process_image(img_element, None))
            case "figcaption":
                figcaption = self._process_children(element, level, is_pre=is_pre).replace(LB_SYMBOL, "\n").strip()
                figcaptions = figcaption.replace("\n\n", "\n").split("\n")
                parts.append("\n".join([f"*{caption}*" for caption in figcaptions]))
            case "table":
                parts.append(self._process_table(element, level))
            case "math":  # 处理latex公式
                semantics = filter_tag(element.find("semantics"))
                if semantics:
                    tex = filter_tag(semantics.find(attrs={"encoding": "application/x-tex"}))
                    if tex:
                        parts.append(f"$$ {tex.text} $$")
            case "script":  # 处理github gist
                parts.append(self._process_gist(element))
            case _:
                parts.append(self._process_children(element, level, is_pre=is_pre))
        result = "".join(parts)
        if result and is_block_element(element.name):
            if not element.children or not is_pure_block_children(element):
                result = f"{LB_SYMBOL}{result}{LB_SYMBOL}"
        return result

    def _process_children(self, element: Tag, level: int = 0, is_pre: bool = False) -> str:
        parts = []
        if element.children:
            # new_level = level + 1 if element.name in HtmlMarkdownParser.TRUSTED_ELEMENTS else level
            for child in element.children:
                if isinstance(child, NavigableString):
                    if is_pre:
                        parts.append(child)
                    else:
                        result = collapse_spaces(child).replace("<", "&lt;").replace(">", "&gt;")
                        if result.strip():
                            parts.append(result)
                        # print(element.name, level, result)
                elif isinstance(child, Tag):
                    result = self._process_element(child, level, is_pre=is_pre)
                    if is_pre:
                        parts.append(result)
                    elif result.strip():
                        parts.append(result.strip())
        return "".join(parts) if is_pre or level > 0 else "".join(parts).strip()

    def _process_list(self, element: Tag, level: int) -> str:
        indent = "  " * level
        child_list = element.find_all(recursive=False)
        is_ol = element.name == "ol"
        parts = []
        for i, child in enumerate(child_list):
            child = filter_tag(child)
            if child:
                if child.name == "li":
                    content = self._process_children(child, level).replace(LB_SYMBOL, "\n").strip()
                    if content:  # 忽略空内容
                        prefix = f"{i + 1}." if is_ol else "-"
                        parts.append(f"{indent}{prefix} {content}")
                elif child.name == "ul" or child.name == "ol":
                    content = self._process_element(child, level + 1)
                    if content:  # 忽略空内容
                        parts.append(f"{content.replace(LB_SYMBOL, '\n')}")
        if not parts:
            return ""  # 所有内容都为空则返回空字符串
        return "\n".join(parts)

    def _process_codeblock(self, element: Tag, level: int) -> str:
        # 找出所有 code 标签（可能为 0 个、1 个或多个）
        code_elements = element.find_all("code") or [element]

        # 处理每一个 code 标签并拼接
        code_parts = [
            self._process_children(code_el, level, is_pre=True).replace(LB_SYMBOL, "\n")
            for code_el in code_elements
            if isinstance(code_el, Tag)
        ]
        code = "\n".join(code_parts).strip()

        if is_sequentially_increasing(code):
            return ""  # 忽略行号

        # 尝试提取语言：从第一个 code 标签的 class 中提取 language
        first_code_el = code_elements[0]
        language = (
            next((cls.split("-")[1] for cls in (first_code_el.get("class") or []) if cls.startswith("language-")), "")
            if isinstance(first_code_el, Tag)
            else ""
        )
        if not language:
            language = detect_language(None, code)
        return f"```{language}\n{code}\n```" if language else f"```\n{code}\n```"

    def _process_table(self, element: Tag, level: int) -> str:
        if element.find("pre"):
            return self._process_children(element, level)
        # 获取所有行，包括 thead 和 tbody
        rows = element.find_all("tr")
        if not rows:
            return ""
        # 解析表头（如果有）
        headers = []
        first_row = filter_tag(rows.pop(0))
        if first_row and first_row.find("th"):
            headers = [th.get_text(strip=True) for th in first_row.find_all("th")]
        # 解析表身
        body = [[td.get_text(strip=True) for td in row.find_all("td")] for row in rows if isinstance(row, Tag)]
        # 处理缺失的表头
        if not headers and body:
            headers = body.pop(0)
        # 统一列数
        col_count = max(len(headers), max((len(row) for row in body), default=0))
        headers += [""] * (col_count - len(headers))
        for row in body:
            row += [""] * (col_count - len(row))
        # 生成 Markdown 表格
        markdown_table = []
        markdown_table.append("| " + " | ".join(headers) + " |")
        markdown_table.append("|-" + "-|-".join(["-" * len(h) for h in headers]) + "-|")
        for row in body:
            markdown_table.append("| " + " | ".join(row) + " |")
        return "\n".join(markdown_table)

    def _process_image(self, element: Tag, source: Tag | None) -> str:
        src = (
            get_attr_text(element.attrs.get("src"))
            if source is None
            else get_attr_text(source.attrs.get("srcset")).split()[0]
        )
        alt = get_attr_text(element.attrs.get("alt"))
        if src:
            if not src.startswith("http") and self.article.url:
                src = urljoin(self.article.url, src)
            return f"![{alt}]({src})"
        return ""

    def _process_gist(self, element: Tag) -> str:
        src = get_attr_text(element.attrs.get("src"))
        pattern = r"/([0-9a-f]+)(?:\.js)?$"
        match = re.search(pattern, src)
        if match:
            gist_id = match.group(1)
            url = f"https://api.github.com/gists/{gist_id}"
            response = requests.get(url)
            response.encoding = "utf-8"
            if response.status_code == 200:
                data = response.json()
                gists = []
                for filename, info in data["files"].items():
                    code = info["content"]
                    language = detect_language(filename, code)
                    gists.append(f"```{language}\n{code}\n```")
                return "\n\n".join(gists)
            else:
                print(f"Fetch gist error: {response.status_code}")
        return ""


def is_block_element(element_name: str) -> bool:
    return element_name in BLOCK_ELEMENTS


def is_pure_block_children(element: Tag) -> bool:
    for child in element.children:
        if isinstance(child, NavigableString):
            if child.strip():  # 有非空文本
                return False
        elif isinstance(child, Tag) and not is_block_element(child.name):
            return False
    return True



from typing import override
from bs4 import BeautifulSoup
from bs4.element import Tag

from ..extractor import Extractor
from ..utils import filter_tag, get_attr_text

class ToutiaoExtractor(Extractor):
    """
    今日头条
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        title_tag = soup.title
        title = title_tag.get_text(strip=True) if title_tag else None
        return title is not None and title.endswith(" - 今日头条")

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "article-content"})

    @override
    def extract_img(self, element: Tag) -> Tag:
        img_els = element.find_all("img")
        for img_el in img_els:
            img_tag = filter_tag(img_el)
            if img_tag:
                src = get_attr_text(img_tag.attrs.get("data-src"))
                if src:
                    img_tag.attrs["src"] = src
        return element



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_site_name

class SspaiExtractor(Extractor):
    """
    少数派
    """
    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend(
            [
                lambda el: "class" in el.attrs and "comment__list" in el.attrs["class"],
                lambda el: "class" in el.attrs and "comment__footer__wrapper" in el.attrs["class"],
            ]
        )

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_site_name(soup) == "少数派 - 高品质数字消费指南"

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "article__main__wrapper"})



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import is_matched_canonical

class CnBlogsExtractor(Extractor):
    """
    博客园
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend(
            [
                lambda el: "id" in el.attrs and "blog_post_info_block" in el.attrs["id"],
                lambda el: "class" in el.attrs and "postDesc" in el.attrs["class"],
            ]
        )

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return is_matched_canonical("https://www.cnblogs.com", soup)

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "post"})

    @override
    def extract_description(self, soup: BeautifulSoup) -> str:
        return ""



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor

class OsChinaExtractor(Extractor):
    """
    开源中国
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        title_tag = soup.title
        title = title_tag.get_text(strip=True) if title_tag else None
        return title is not None and title.endswith(" - OSCHINA - 中文开源技术交流社区")

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "detail-box"})



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_og_site_name


class JetbrainsBlogExtractor(Extractor):
    """
    blog.jetbrains.com
    """

    def __init__(self):
        super().__init__()
        self.attrs_to_clean.extend(
            [
                lambda el: "class" in el.attrs and "content__row" in el.attrs["class"],
                lambda el: "class" in el.attrs and "content__pagination" in el.attrs["class"],
                lambda el: "class" in el.attrs and "content__form" in el.attrs["class"],
                lambda el: "class" in el.attrs and "tag" in el.attrs["class"],
                lambda el: "class" in el.attrs and "author-post" in el.attrs["class"],
            ]
        )

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_og_site_name(soup) == "The JetBrains Blog"

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "content"})



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import filter_tag


class FreediumExtractor(Extractor):
    """
    freedium.cfd
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        title_tag = soup.title
        title = title_tag.get_text(strip=True) if title_tag else None
        return title is not None and title.endswith(" - Freedium")

    @override
    def article_container(self) -> tuple:
        return ("div", {"class": "main-content"})

    @override
    def extract_title(self, soup: BeautifulSoup) -> str:
        title_tag = filter_tag(soup.find("h1"))
        if title_tag:
            title = title_tag.get_text(strip=True)
            title_tag.decompose()
            return title
        return super().extract_title(soup)

    @override
    def extract_description(self, soup: BeautifulSoup) -> str:
        description_tag = soup.find("h2")
        if description_tag:
            description = description_tag.get_text(strip=True)
            description_tag.decompose()
            return description
        return super().extract_description(soup)



from typing import override
from bs4 import BeautifulSoup

from ..extractor import Extractor
from ..utils import get_title


class ClaudeDocExtractor(Extractor):
    """
    Anthropic
    """

    @override
    def can_handle(self, soup: BeautifulSoup) -> bool:
        return get_title(soup).endswith(" \\ Anthropic")

    @override
    def article_container(self) -> tuple:
        return ("article", None)

    @override
    def extract_url(self, soup: BeautifulSoup) -> str:
        return "https://www.anthropic.com/"
